# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
from time import sleep
from typing import Optional

import psycopg2 as pg
import requests
from pycapo import CapoConfig
from sqlalchemy.engine import Connection

logger = logging.getLogger("ws_annihilator")


class MDDBConnector:
    """Use this connection to question the archive"""

    def __init__(self):
        self.connection = self._connect_to_db()

    @staticmethod
    def _connect_to_db() -> Connection:
        """
        Establish a DB connection

        :return:
        """
        settings = CapoConfig().settings("metadataDatabase")

        host_slash_db = settings["jdbcUrl"][len("jdbc:postgresql://") :]
        host, dbname = host_slash_db.split("/")
        port = 5432
        if ":" in host:
            host, port = host.split(":")
        try:
            conn = pg.connect(
                host=host,
                port=port,
                database=dbname,
                user=settings.jdbcUsername,
                password=settings.jdbcPassword,
            )
            return conn
        except Exception as exc:
            logger.error(f"Unable to connect to database: {exc}")
            raise exc

    def cursor(self):
        return self.connection.cursor()

    def close(self):
        self.connection.close()

    def commit(self):
        self.connection.commit()


def send_notification(notification_url: str, template: str, parameters: Optional[dict] = None) -> None:
    """Send an email notification through the notification service

    :param notification_url: URL of the notification service to use
    :param template: Name of mustache template for email, must correspond to entry in the notification_templates table
    :param parameters: Parameters for given template
    :raises RuntimeError: When the notification service fails to send the email
    """
    if parameters is None:
        parameters = dict()
    url = f"{notification_url}/notify/{template}/send"
    response = requests.post(url, json=parameters)
    if response.status_code != 200:
        # wait 5 seconds and try again:  There is an issue with message collisions in the notification system,
        # this should be unnecessary once it it resolved.
        sleep(5)
        response = requests.post(url, json=parameters)
        if response.status_code != 200:
            raise RuntimeError(
                f"Notification failure: Expected 200 (OK) response, got {response.status_code} response instead: {response.text}"
            )
    try:
        json = response.json()
    except requests.JSONDecodeError:
        raise RuntimeError(f"Notification failure: Expected JSON response, instead got: {response.text}")
    if json["errors"] or json["message"] != "Email sent":
        raise RuntimeError(f"Notification failure: Expected response with no errors, instead got: {json}")
