#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import json
from typing import Dict, List, Union

from workspaces.workflow.schema import (
    Workflow,
    WorkflowRequest,
    WorkflowRequestFile,
    WorkflowTemplate,
    ProcessingSite,
    Controller,
    WorkflowProgress,
)
from workspaces.workflow.services.interfaces import WorkflowInfoIF


class FakeWorkflowInfo(WorkflowInfoIF):
    """
    A fake workflow info. Providing this as the substrate to other workflow
    services for testing will enable us to ensure that the right things are
    happening under the hood.
    """

    def __init__(self):
        self.requests = []

    def lookup_workflow_request(self, request_id: int) -> Workflow:
        return self.requests[request_id - 1]

    def all_workflows(self) -> List[Workflow]:
        return [self.lookup_workflow_definition("null")]

    def lookup_processing_site(self, site_key: str) -> ProcessingSite:
        if site_key == "NAASC":
            return ProcessingSite(processing_site="NAASC", template_key="remote-cv", data_center=True)
        else:
            return NotImplementedError

    def lookup_controller(self, controller_key: str) -> Controller:
        if controller_key == "WS":
            return Controller(controller_key="WS", message_handler="default", routing_key_prefix="")
        else:
            return NotImplementedError

    def lookup_workflow_stage(self, request_id: int, stage: str) -> Union[WorkflowProgress, None]:
        return NotImplementedError

    def all_processing_sites(self) -> List[ProcessingSite]:
        return [self.lookup_processing_site("NAASC")]

    def all_controllers(self) -> List[Controller]:
        return [self.lookup_controller("WS")]

    def all_external_controllers(self) -> List[Controller]:
        return [self.lookup_controller("AAT"), self.lookup_controller("VLASS")]

    def create_workflow_request(self, workflow: Union[str, Workflow], argument: Dict) -> WorkflowRequest:
        workflow_name = workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow
        request = WorkflowRequest(workflow_name=workflow_name, argument=argument)
        self.save_request(request)
        return request

    def save_request(self, request: WorkflowRequest):
        request.workflow_request_id = len(self.requests) + 1
        self.requests.append(request)

    def save_file(self, request: WorkflowRequest, filename: str, content: bytes) -> WorkflowRequestFile:
        request.files.append(WorkflowRequestFile(filename=filename, content=content))

    def lookup_workflow_definition(self, name: str) -> "Workflow":
        if name == "null":
            w = Workflow(workflow_name="null")
            w.templates.append(
                WorkflowTemplate(
                    filename="null.condor",
                    content=b'executable = null\narguments = "-n"\nlog = condor.log\n\nqueue',
                )
            )
            return w
        else:
            return NotImplementedError


info = FakeWorkflowInfo()


def test_workflow_creation():
    # FIXME: This is a unit test, but it seems to only be testing this mocked FakeWorkflowInfo object and not be testing
    #  the actual system in any meaningful way
    # This seems like a bit of a blunt instrument for testing the service. We should probably break the service down
    # into smaller pieces.
    global info

    # create request
    request = info.create_workflow_request(info.lookup_workflow_definition("null"), json.loads('{"argument":"-n"}'))

    # did the workflow info receive a new workflow request?
    assert request in info.requests
