#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Workspaces Workflow Service - Workflow Recovery Facility
"""
import logging
import os
import pathlib
from pathlib import Path
from typing import List

import pendulum
from sqlalchemy import text

from workspaces.workflow.schema import WorkflowRequest
from workspaces.workflow.services.interfaces import WorkflowInfoIF
from workspaces.workflow.services.monitor import WorkflowMonitor

logger = logging.getLogger(__name__)


def _ensure_initial_startup() -> bool:
    """
    Docker containers are recreated whenever software is deployed or containers are restarted.
    Check if initialization file is present in container to ensure that workflow reacquisition
    only runs on container restart.

    :return: True if file doesn't exist, False otherwise
    """

    ensure_initial_file = "reacquire.txt"

    if pathlib.Path(ensure_initial_file).exists():
        return False

    content = f"The Workflow Reacquire system was started within this container at {pendulum.now().utcnow()}\n"

    eif = open(ensure_initial_file, "w")
    eif.write(content)
    eif.close()

    logger.info("Workflow reacquisition beginning...")
    return True


def get_environment() -> str:
    """
    Determine which environment the system is running on. Important for making sure only
    requests belonging to the current environment reacquire monitor processes.

    :return: string representing the environment - i.e. dev, test, or prod
    """
    capo_profile = os.environ["CAPO_PROFILE"]

    if capo_profile != "docker":
        environment = capo_profile.split("-")
        return environment[1]

    return capo_profile


class MonitorRecover:
    def __init__(self, info: WorkflowInfoIF):
        self.info = info

    def _on_restart(self) -> List[WorkflowRequest]:
        """
        Mine workflow_requests table for in-flight requests

        :return: List of in-flight requests eligible for reacquisition
        """

        inflights = []
        profile = get_environment()

        for row in self.info.session.get_bind().execute(
            text(
                f"""
                SELECT workflow_request_id as wf_id
                FROM workflow_requests
                WHERE state not in ('Created', 'Complete', 'Failed', 'Error')
                AND results_dir like ('%{profile}%')
                AND cleaned = 'False'
                """
            )
        ):
            inflights.append(self.info.lookup_workflow_request(request_id=row[0]))

        return inflights

    @staticmethod
    def _reacquire_logfile(wf_request: WorkflowRequest) -> Path:
        """
        For an In-flight WorkflowRequest, reacquire the condor log file

        :return: Path to the HTCondor logfile
        """
        results_dir = Path(wf_request.results_dir)

        if results_dir.is_dir():
            # because for some reason these are not the same....
            if results_dir == pathlib.Path(".") or results_dir == pathlib.Path.cwd() or results_dir == pathlib.Path(""):
                logger.info(f"Results directory for Request #{wf_request.workflow_request_id} not found. Skipping.")
                return Path(".")
            else:
                logger.info(f"Searching for log file for Request #{wf_request.workflow_request_id}...")
                dag_files = list(results_dir.glob("*.dag"))
                if dag_files:
                    logger.info("Found dagman logfile!")
                    return Path(str(dag_files[0]) + ".dagman.log")

                for root, dirs, files in os.walk(results_dir):
                    for file in files:
                        if file.endswith(".log") and (
                            file.startswith(wf_request.workflow_name) or file.startswith("condor")
                        ):
                            logger.info("Found condor logfile!")
                            return file
        else:
            logger.info(
                f"Results directory {results_dir} for request #{wf_request.workflow_request_id} "
                f"is not a directory! Skipping."
            )
            # signal not a directory with current location
            return Path(".")

    @staticmethod
    def _restart_monitor(logfile: Path, wf_request: WorkflowRequest):
        """
        Restart a wf_monitor process for the provided WorkflowRequest

        :param logfile: The HTCondor logfile to monitor
        :param wf_request: The WorkflowRequest to update
        :return:
        """
        # because for some reason these are not the same....
        if logfile == Path.cwd() or logfile == Path(".") or logfile is None:
            return

        logger.info("Running wf_monitor.")
        is_dag = True if str(logfile).endswith(".dagman.log") else False

        WorkflowMonitor(logfile, wf_request, is_dag).start()
        logger.info(f"Successfully reacquired monitor for request #{wf_request.workflow_request_id}. Continuing...")

    def reacquire_inflight_requests(self):
        """
        Run reacquire for all in-flight requests

        :return:
        """

        if _ensure_initial_startup():
            inflight = self._on_restart()

            for request in inflight:
                logger.info(f"Reacquiring monitor for Request #{request.workflow_request_id}...")
                self._restart_monitor(self._reacquire_logfile(request), request)
        else:
            logger.info("Not initial container startup, skipping workflow reacquisition.")
